//This example shows how to animate graphics on a VGA screen. No backbuffering is used... just try it.
//You need to connect a VGA screen cable and an external DAC (simple R2R does the job) to the pins specified below.
//cc by-sa 4.0 license
//bitluni

// modified demo to animate replicating Langton's Ants

#include <ESP32Lib.h>
#include <Ressources/Font6x8.h>
#include <math.h>

//pin configuration
const int redPins[] = {2, 4, 12, 13, 14};
const int greenPins[] = {15, 16, 17, 18, 19};
const int bluePins[] = {21, 22, 23, 27};
const int hsyncPin = 32;
const int vsyncPin = 33;

//VGA Device
VGA14Bit vga;

// Maximum number of Langton's ants
#define NANTS 4096 

// x y location of each ant
uint8_t ax[NANTS];
uint8_t ay[NANTS];

// direction ant is moving
uint8_t ad[NANTS];

// number of ants
uint16_t an;

long lastTime = 0;

//initial setup
void setup()
{
	//initializing i2s vga (with only one framebuffer)
	vga.init(vga.MODE200x150, redPins, greenPins, bluePins, hsyncPin, vsyncPin);
	//setting the font
	vga.setFont(Font6x8);

  // 2___
  vga.fillEllipse(25, 50, 23, 48, 0x3fff);
  vga.fillEllipse(25, 50, 7, 20, 0);
  vga.fillRect(2, 49, 48, 49, 0);
  for (int x = 0; x<16; ++x) {
    vga.line(2+x,114,x+32, 49, 0x3fff);
  }
  vga.fillRect(2,114,46,26, 0x3fff);

  // _0__
  vga.fillEllipse(75,72,23,68, 0x3fff);
  vga.fillEllipse(75,72,8,40, 0);

  // __2_
  vga.fillEllipse(125, 50, 23, 48, 0x3fff);
  vga.fillEllipse(125, 50, 7, 20, 0);
  vga.fillRect(102, 49, 48, 49, 0);
  for (int x = 0; x<16; ++x) {
    vga.line(102+x,114,x+132, 49, 0x3fff);
  }
  vga.fillRect(102,114,46,26, 0x3fff);

  // ___0
  vga.fillEllipse(175,72,23,68, 0x3fff);
  vga.fillEllipse(175,72,8,40, 0);

  an = 0;
}

int adx[4] = {-1, 0, 1, 0}; 
int ady[4] = {0, -1, 0, 1}; 
int steps = 1;

// ant colours
uint16_t colour[12] = {0x1f, 0x3e0, 0x3c00, 0x3ff, 0x3fe0, 0x3c1f, 0x3ef, 0x1ff, 0x3de0, 0x1fe0, 0x3c0f, 0x1c1f};

//the loop is done every frame
void loop()
{
	//setting the text cursor to the lower left corner of the screen
	vga.setCursor(0, vga.yres - 8);
	//setting the text color to white with opaque black background
	vga.setTextColor(vga.RGB(0xffffff), vga.RGBA(0, 0, 0, 255));
	//printing the fps
	vga.print("fps: ");
	static long f = 0;
  long t = millis();
	vga.print(long((f++ * 1000) / t));


  if (t > (lastTime + 10)) {
    lastTime = t;
    uint16_t an0 = an;
    if ((an == 0) && (steps == 0)) {
      // initialise first ant
      an = 1;
      ax[0] = 100;
      ay[0] = 75;
      ad[0] = 0;
    }
    for (uint16_t ai = 0; ai < an0; ++ai) {
      int x = ax[ai];
      int y = ay[ai];
      int dx = adx[ad[ai]];
      int dy = ady[ad[ai]];
      x += dx;
      if (x < 0) x = 199;
      if (x >= 200) x = 0;
      y += dy;
      if (y < 0) y = 140;
      if (y >= 141) y = 0;
  
      ax[ai] = x;
      ay[ai] = y;
  
      uint16_t c = vga.get(x, y);
      if (c == 0) {
        // turn clockwise
        vga.dotFast(x, y, colour[ai%12]);
        ad[ai] = (ad[ai] + 1)&3;
      } else {
        // turn counterclockwise
        ad[ai] = (ad[ai] + 3)&3;
        vga.dotFast(x, y, 0);
      }
  
      if ((an < NANTS) && (steps == 0)) {
        // replicate ant
        ax[an] = x;
        ay[an] = y;
        ad[an] = (ad[ai] + 2)&3;
        ++an;
      }
    }
    ++steps;
    if (steps > ((an==0) ? 500 : 267)) steps = 0;
    vga.setCursor(60, vga.yres - 8);
    vga.print("Ants: ");
    vga.print(an);    
  }
}
